#include <qapplication.h>
#include <qbitmap.h>
#include <qfiledialog.h>
#include <qlayout.h>
#include <qmainwindow.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>
#include <qpushbutton.h>
#include <qwidget.h>
#include <zlib.h>

class Editor : public QWidget
{
Q_OBJECT
	QWidget *w;
	QPushButton *siatka[32][16];
	QGridLayout *grid;
	QPopupMenu *file, *help;
	unsigned char d[64];
	QString dir;
	public:
	QMainWindow *window;
	Editor();
	void setChar(unsigned int width, unsigned int height, unsigned char data[]);
	unsigned char *getData(unsigned int width, unsigned int height);
	public slots:
	void open();
	void end();
	void about();
	void aboutQt();
};

class Font : public QObject
{
Q_OBJECT
	QMainWindow *window;
	QWidget *w;
	QPushButton *siatka[512];
	QGridLayout *grid;
	QPopupMenu *file;
	QPopupMenu *help;
	QString dir;
	void changePixmap(unsigned int n);
	public:
	class PSF
	{
		public:
		PSF():type(0) {};
		unsigned int type;
		unsigned int unicodeLength;
		unsigned int restLength;
		unsigned char *data;
		unsigned char *unicodeData;
		unsigned char *rest;
		unsigned char header[4];
		unsigned char header2[28];
		unsigned int version, headersize, flags, length, charsize, height, width;
		PSF(const char *name);
		~PSF();
		void setData(unsigned int n, unsigned char d[16]);
		unsigned char *getData(unsigned int n);
	};
	PSF *f;
	Font(const char *name);
	protected:
	bool eventFilter( QObject *obj, QEvent *ev );
	public slots:
	void save();
	void close();
	void info();
};

Editor *edi;

Editor::Editor() : dir("/lib/kbd/consolefonts")
{
	window = new QMainWindow;
	w = new QWidget(window);
	file = new QPopupMenu(window);
	help = new QPopupMenu(window);
	window->menuBar()->insertItem("&File", file);
	window->menuBar()->insertItem("&Help", help);
	file->insertItem("&Open", this, SLOT(open()));
	file->insertItem("&Exit", this, SLOT(end()));
	help->insertItem("&About", this, SLOT(about()));
	help->insertItem("About &Qt", this, SLOT(aboutQt()));
	grid = new QGridLayout(w, 32, 32);
	for (int i = 0; i < 32; i++)
		for (int j = 0; j < 16; j++) {
			siatka[i][j] = new QPushButton(w);
			siatka[i][j]->setToggleButton(true);
			siatka[i][j]->setMaximumSize(20, 20);
			grid->addWidget(siatka[i][j], i, j);
		}
	window->setCentralWidget(w);
	window->show();
}

void
Editor::open()
{
	QString name = QFileDialog::getOpenFileName(dir);
	new Font(name);
}

void
Editor::end()
{
	QApplication::exit();
}

void
Editor::about()
{	
	QMessageBox::about(window, "About AFE", "PSF font editor\n(C) 2007 Witold Filipczyk");
}

void
Editor::aboutQt()
{
	QMessageBox::aboutQt(window);
}

void
Editor::setChar(unsigned int width, unsigned int height, unsigned char data[])
{
	unsigned int charsize = (width + 7) / 8;
	for (unsigned int i = 0; i < height; i++)
		for (unsigned int j = 0; j < charsize; j++) {
			unsigned int ind = i * charsize + j;
			for (int bit = 7; bit >= 0; bit--) {
				if (data[ind] & (1 << bit))
					siatka[i][7 - bit + j * 8]->setOn(true);
				else
					siatka[i][7 - bit + j * 8]->setOn(false);
			}
		}
}

unsigned char *
Editor::getData(unsigned int width, unsigned int height)
{
	for (unsigned int i = 0; i < 64; i++)
		d[i] = 0;
	
	unsigned int charsize = (width + 7) / 8;
	for (unsigned int i = 0; i < height; i++)
		for (unsigned int j = 0; j < charsize; j++) {
			unsigned int ind = i * charsize + j;
			for (int bit = 7; bit >= 0; bit--) {
				if (siatka[i][7 - bit + j * 8]->isOn())
					d[ind] |= (1 << bit);
			}
		}
	return d;
}

unsigned int
calc(unsigned char data[4])
{
	return data[0] + (data[1] << 8) + (data[2] << 16) + (data[3] << 24);
}

Font::PSF::PSF(const char *name):type(0), data(NULL), unicodeData(NULL), rest(NULL)
{
	gzFile gz = gzopen(name, "rb");
	int l = 0;

	if (!gz)
		return;
	gzread(gz, header, 4);
	if (header[0] == 0x36 && header[1] == 0x04) {	
		if (header[2] & 1)
			length = 512;
		else
			length = 256;
		if (header[2] & 2)
			l = 1;
		height = charsize = header[3];
		data = (unsigned char *)malloc(length * charsize);
		if (!data) {
			gzclose(gz);
			return;
		}
		width = 8;
		data = (unsigned char *)malloc(length * charsize);
		if (!data) {
			gzclose(gz);
			return;
		}
		gzread(gz, data, length * charsize);
		type = 1;
	} else if (header[0] == 0x72 && header[1] == 0xb5 && header[2] == 0x4a && header[3] == 0x86) {
		gzread(gz, header2, 28);
		version = calc(header2);
		headersize = calc(header2 + 4);
		flags = calc(header2 + 8);
		length = calc(header2 + 12);
		charsize = calc(header2 + 16);
		height = calc(header2 + 20);
		width = calc(header2 + 24);
		l = flags & 1;
		restLength = headersize - 32;
		if (restLength > 0) {
			rest = (unsigned char *)malloc(restLength);
			if (!rest) {
				gzclose(gz);
				return;
			}
			gzread(gz, rest, restLength);
		}
		data = (unsigned char *)malloc(length * charsize);
		if (!data) {
			gzclose(gz);
			return;
		}
		gzread(gz, data, length * charsize);
		type = 2;
	}
	unicodeLength = 0;
	while (l) {
		unicodeData = (unsigned char *)realloc(unicodeData, unicodeLength + 65536);
		l = gzread(gz, unicodeData + unicodeLength, 65536);
		if (l <= 0) break;
		unicodeLength += l;
	}
	gzclose(gz);
}

void
Font::PSF::setData(unsigned int n, unsigned char d[16])
{
	unsigned int ind = n * charsize;

	for (unsigned int i = 0; i < charsize; i++)
		data[ind + i] = d[i];
}

unsigned char *
Font::PSF::getData(unsigned int n)
{
	return data + (n * charsize);
}

Font::PSF::~PSF()
{
	if (type) {
		if (data)
			free(data);
		if (unicodeData)
			free(unicodeData);
		if (rest)
			free(rest);
	}
}

void
Font::changePixmap(unsigned int n)
{
	QBitmap q(f->width, f->height, f->getData(n));
	siatka[n]->setPixmap(q);
} 

Font::Font(const char *name) : dir(getenv("HOME"))
{
	char n[4];
	f = new PSF(name);
	if (!f->type)
		return;
	window = new QMainWindow;
	file = new QPopupMenu(window);
	help = new QPopupMenu(window);
	window->menuBar()->insertItem("&File", file);
	window->menuBar()->insertItem("&Help", help);
	file->insertItem("&Save as", this, SLOT(save()));
	file->insertItem("&Close", this, SLOT(close()));
	help->insertItem("&Help", this, SLOT(info()));
	w = new QWidget(window);
	window->setCaption(name);
	grid = new QGridLayout(w, f->length / 32, 32);
	for (unsigned int i = 0; i < f->length; i++) {
		unsigned int ii = i / 32;
		unsigned int j = i % 32;
		snprintf(n, 4, "%d", i);
		siatka[i] = new QPushButton(w, n);
		changePixmap(i);
		siatka[i]->installEventFilter(this);
		grid->addWidget(siatka[i], ii, j);
	}
	window->setCentralWidget(w);
	window->show();
}

bool
Font::eventFilter( QObject *obj, QEvent *ev)
{
	if (ev->type() == QEvent::MouseButtonPress) {
		int number = atoi(obj->name());
		QMouseEvent *me = (QMouseEvent *)ev;
		ButtonState b = me->button();
		unsigned char *d;

		switch (b) {
		case Qt::LeftButton:
			edi->setChar(f->width, f->height, f->getData(number));
			break;
		case Qt::MidButton:
			f->setData(number, edi->getData(f->width, f->height));
			changePixmap(number);
			break;
		case Qt::RightButton:
			d = edi->getData(f->width, f->height);
			edi->setChar(f->width, f->height, f->getData(number));
			f->setData(number, d);
			changePixmap(number);
		default:
			break;
		}
		return true;
	} else {
		return false;
	}
	
}

void
Font::save()
{
	QString filename = QFileDialog::getSaveFileName(dir);
	if (filename.endsWith(".gz")) {
		gzFile gz = gzopen(filename, "wb");
		if (!gz) {
			return;
		}
		if (f->type == 1) {
			gzwrite(gz, f->header, 4);
			gzwrite(gz, f->data, (f->charsize * f->length));
			gzwrite(gz, f->unicodeData, f->unicodeLength);
		} else if (f->type == 2) {
			gzwrite(gz, f->header, 4);
			gzwrite(gz, f->header2, 28);
			gzwrite(gz, f->rest, f->restLength);
			gzwrite(gz, f->data, (f->charsize * f->length));
			gzwrite(gz, f->unicodeData, f->unicodeLength);
		}
		gzclose(gz);
	} else {
		FILE *fi = fopen(filename, "wb");
		if (!fi)
			return;
		if (f->type == 1) {
			fwrite(f->header, 1, 4, fi);
			fwrite(f->data, 1, (f->charsize * f->length), fi);
			fwrite(f->unicodeData, 1, f->unicodeLength, fi);
		} else if (f->type == 2) {
			fwrite(f->header, 1, 4, fi);
			fwrite(f->header2, 1, 28, fi);
			fwrite(f->rest, 1, f->restLength, fi);
			fwrite(f->data, 1, (f->charsize * f->length), fi);
			fwrite(f->unicodeData, 1, f->unicodeLength, fi);
		}
		fclose(fi);
	}
}

void
Font::close()
{
	window->hide();
}

void
Font::info()
{
	QMessageBox::information(window, "Help", "Left mouse button: edit\n"
	"Middle mouse button: paste\nRight mouse button: swap", QMessageBox::Ok);	
}

int main(int argc, char *argv[])
{
	QApplication app(argc, argv);
	edi = new Editor;
	for (int i = 0; i < argc; i++)
		new Font(argv[i]);
	return app.exec();
}

#include "afe.moc"
